#include <math.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <omp.h>

#define NN 4096
#define NM 4096

#define IDX(i,j,m) ((i)*(m)+(j))

int main(int argc, char** argv)
{
  const int n = NN;
  const int m = NM;
  const int niter_max = 1000;

  const double tol = 1.0e-6;
  
  double error = 1.0;
  int niter = 0;

  double*    A = (double*)malloc(n * m * sizeof(double));
  double* Anew = (double*)malloc(n * m * sizeof(double));

  memset(A, 0, n * m * sizeof(double));
  memset(Anew, 0, n * m * sizeof(double));

  for (int j = 0; j < n; ++j) {
       A[j*m] = 1.0;
    Anew[j*m] = 1.0;
  }

  printf("Jacobi relaxation Calculation: %d x %d mesh\n", n, m);
  
  double start = omp_get_wtime();
    
  #pragma omp target data map(to: Anew[0:n*m]) map(A[0:n*m])
  while (error > tol && niter < niter_max) {
    error = 0.0;

    #pragma omp target teams distribute parallel for reduction(max:error)
    for (int j = 1; j < n-1; ++j) {
      for (int i = 1; i < m-1; ++i) {
        unsigned int idx = i + j * m; 
        Anew[idx] = 0.25 * ( A[idx+1] + A[idx-1] + A[idx-m] + A[idx+m]);
        error = fmax(error, fabs(Anew[idx] - A[idx]));
      }
    }

    #pragma omp target teams distribute parallel for
    for (int j = 1; j < n-1; ++j) {
      for (int i = 1; i < m-1; i++ ) {
        A[i + j * m] = Anew[i + j * m];
      }
    }

    if(niter % 100 == 0) printf("%5d, %0.6f\n", niter, error);

    niter++;
  }

  double elapsed = omp_get_wtime() - start;
  printf(" Elapsed time: %f s\n", elapsed);

  free(A);
  free(Anew);

  return 0;
}