    program laplace
      use omp_lib
    
      implicit none
    
      integer, parameter :: n = 4096, m = 4096, niter_max = 1000
    
      integer :: i, j, niter
      double precision, dimension (:,:), allocatable :: A, Anew
    
      double precision :: tol = 1.0E-6, error = 1.0d0
      double precision :: tstart, tstop
    
      allocate ( A(0:n-1,0:m-1), Anew(0:n-1,0:m-1) )
    
      A    = 0.0d0
      Anew = 0.0d0
    
      A(0,:)    = 1.0d0
      Anew(0,:) = 1.0d0
       
      write(*,'(a,i5,a,i5,a)') 'Jacobi relaxation Calculation:', n, ' x', m, ' mesh'
     
      tstart = omp_get_wtime()
    
      niter = 0
    
      !$omp target data map(to: Anew) map(A)
      do while (error .gt. tol .and. niter .lt. niter_max)
        error=0.0d0
    
        !$omp target teams distribute parallel do collapse(2) reduction(max:error)
        do j = 1,m-2
          do i = 1,n-2
            Anew(i,j) = 0.25d0 * ( A(i+1,j  ) + A(i-1,j  ) + &
                                   A(i  ,j-1) + A(i  ,j+1) )
            error = max(error, abs(Anew(i,j) - A(i,j)))
          end do
        end do
    
        if (mod(niter, 100) .eq. 0 ) write(*,'(i5,f10.6)') niter, error
        niter = niter + 1
    
        !$omp target teams distribute parallel do collapse(2)
        do j = 1,m-2
          do i = 1,n-2
            A(i,j) = Anew(i,j)
          end do
        end do
      end do
      !$omp end target data
    
      tstop = omp_get_wtime()
      write(*,'(a,f10.3,a)') ' Elapsed: ', tstop - tstart, ' s'
    
      deallocate(A, Anew)
    end program laplace